///////////////////////////////////////////////////////////////////////
// File:        universalambigs.h
// Description: Data for a universal ambigs file that is useful for
//              any language.
// Author:      Ray Smith
// Created:     Mon Mar 18 11:26:00 PDT 2013
//
// (C) Copyright 2013, Google Inc.
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
///////////////////////////////////////////////////////////////////////

#ifndef TESSERACT_CCUTIL_UNIVERSALAMBIGS_H_
#define TESSERACT_CCUTIL_UNIVERSALAMBIGS_H_

namespace tesseract {

extern const char kUniversalAmbigsFile[];
extern const int ksizeofUniversalAmbigsFile;

}  // namespace tesseract

#endif  // TESSERACT_CCUTIL_UNIVERSALAMBIGS_H_
